from .serializable import Serializable
JsonSerializable = Serializable
try:
    from .yaml_serialization import YamlSerializable
except ImportError:
    pass
from .decoding import *
from .encoding import *



import random


def set_seed(seed: int) -> None:
    random.seed(seed)
    try:
        import numpy as np
        np.random.seed(seed)
    except ImportError:
        pass

    try:
        import torch
    except ImportError:
        pass
    else:
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)



import inspect
from collections import OrderedDict
from dataclasses import MISSING, asdict, dataclass, fields, is_dataclass
from logging import getLogger
from pathlib import Path
from typing import IO, Any, ClassVar, Dict, List, Type, TypeVar, Union

import yaml

from .decoding import register_decoding_fn
from .serializable import D, Serializable, from_dict

logger = getLogger(__name__)


class YamlSerializable(Serializable):
    """Convenience class, just sets different `load_fn` and `dump_fn` defaults
    for the `dump`, `dumps`, `load`, `loads` methods of `Serializable`.

    Uses the `yaml.full_load` and `yaml.dump` for loading and dumping.

    Requires the pyyaml package.
    """
    def dump(self, fp: IO[str], dump_fn=yaml.dump, **kwargs) -> None:
        dump_fn(self.to_dict(), fp, **kwargs)

    def dumps(self, dump_fn=yaml.dump, **kwargs) -> str:
        return dump_fn(self.to_dict(), **kwargs)

    @classmethod
    def load(cls: Type[D], path: Union[Path, str, IO[str]], drop_extra_fields: bool=None, load_fn=yaml.full_load, **kwargs) -> D:
        return super().load(path, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)

    @classmethod
    def loads(cls: Type[D], s: str, drop_extra_fields: bool=None, load_fn=yaml.full_load, **kwargs) -> D:
        return super().loads(s, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)

    @classmethod
    def _load(cls: Type[D], fp: IO[str], drop_extra_fields: bool=None, load_fn=yaml.full_load, **kwargs) -> D:
        return super()._load(fp, drop_extra_fields=drop_extra_fields, load_fn=load_fn, **kwargs)



import functools
from dataclasses import Field
from functools import lru_cache, partial
from logging import getLogger
from typing import Any, Callable, Dict, List, Optional, Type, TypeVar

import typing_inspect as tpi

from ..utils import (get_item_type, get_type_arguments, is_dict, is_list,
                     is_optional, is_set, is_tuple, is_union, str2bool)

T = TypeVar("T")

logger = getLogger(__name__)

_new_metavars: Dict[Type[T], Optional[str]] = {
    # the 'primitive' types don't get a 'new' metavar.
    t: t.__name__ for t in [str, float, int, bytes] 
}

def log_results(fn: Callable[[Type], str]):
    @functools.wraps(fn)
    def _wrapped(t: Type) -> str:
        result = fn(t)
        logger.debug(f"Metavar for type {t}: {result}")
        return result
    return _wrapped


@log_results
def get_metavar(t: Type) -> str:
    """ Gets the metavar to be used for that type in help strings.
    
    This is crucial when using a `weird` auto-generated parsing functions for
    things like Union, Optional, Etc etc.
    
    type the type arguments that were passed to `get_parsing_fn` that
    produced the given parsing_fn.

    returns None if the name shouldn't be changed.
    """
    # TODO: Maybe we can create the name for each returned call, a bit like how
    # we dynamically create the parsing function itself?
    new_name: str = getattr(t, "__name__", None)
   
    optional = is_optional(t)

    if t in _new_metavars:
        return _new_metavars[t]
    
    elif is_union(t):
        args = get_type_arguments(t)
        metavars: List[str] = []
        for type_arg in args:
            if type_arg is type(None):
                continue
            metavars.append(get_metavar(type_arg))
        metavar = "|".join(map(str, metavars))
        if optional:
            return f"[{metavar}]"
        return metavar

    elif is_tuple(t):
        args = get_type_arguments(t)
        if not args:
            return get_metavar(Any)
        logger.debug(f"Tuple args: {args}")
        metavars: List[str] = []
        for arg in args:
            if arg is Ellipsis:
                metavars.append(f"[{metavars[-1]}, ...]")
                break
            else:
                metavars.append(get_metavar(arg))
        return " ".join(metavars)
      
    return new_name
